#!/usr/bin/env python

import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
import warnings
warnings.filterwarnings("ignore")
import logging
import ebf
import json
import optparse
from time import perf_counter as clock
import multiprocessing

import GalCraft.modules.SSPLoader as SSPLoader
import GalCraft.modules.Binner as Binner
import GalCraft.modules.SpecGenerator as SpecGenerator
import GalCraft.modules.CubeMaker as CubeMaker
import GalCraft.modules.Cot as Cot
import GalCraft.modules.utils as utils
from GalCraft._version import __version__


def run_GalCraft(CommandOptions):



    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - -   I N I T I A L I Z E   G A L C R A F T - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    t_init = clock()

    # - - - - - - - - - - INITIALIZATION - - - - - - - - - -

    # Initialize cube name and paths
    cube_name = CommandOptions.configName
    setup_cube_name = 'setup_' + cube_name
    configDir, modelDir, templateDir, outputDir = utils.initialize_Dirs(CommandOptions)
    # Load the Configuration file
    with open(configDir + setup_cube_name + '.json', 'r') as f:
        params = json.load(f)
    # Setup the outputs folder
    filepath = outputDir + cube_name + '/'
    if os.path.exists(filepath)==False: os.mkdir(filepath[:-1])
    # Setup logger
    if params['other_params']['mode'] == 'continue': # For continue mode, change the logger mode from 'w' to 'a'
        params['other_params']['log_mode'] = 'a'
    utils.setupLogfile(logfile=filepath + 'outputs.log', __version__=__version__, mode=params['other_params']['log_mode'], welcome=True)
    logging.info('Loaded the setup file (path %s)' % (configDir + setup_cube_name + '.json'))
    logging.info('Outputs DIR: %s' % filepath)
    # Setup cpu cores to use
    if params['other_params']['ncpu'] == None:
        # n_cores = len(os.sched_getaffinity(0))
        n_cores = multiprocessing.cpu_count()
        logging.info('Change the CPU cores from %s to %s due to the CPU availability.' % (params['other_params']['ncpu'], n_cores) )
        params['other_params']['ncpu'] = n_cores
    else:
        logging.info('Run the framework using %s CPU cores.' % params['other_params']['ncpu'])



    # - - - - - - - - - - INSTRUMENT SETUP - - - - - - - - - -

    # Obtain ./instrument/ dir
    this_dir, this_filename = os.path.split(__file__)
    instrumentDir = os.path.join(this_dir, "instrument") + '/'
    # Load instrument information
    # If "instrument"=="MUSE-WFM", it will load "MUSE-WFM.json" and use the "spatial_resolution" and "spatial_bin";
    # If "instrument"=="DIY", you will design your own instrument using "spatial_resolution" and "spatial_bin" in the configuration file;
    # If "instrument"=="DEFAULT", it will generate a huge cube using all the particles and only will be used in the configuration file.
    inst = params['cube_params']['instrument']
    if inst.upper() != 'DIY' and inst.upper()!='DEFAULT':
        with open(instrumentDir + inst + '.json', 'r') as f:
            inst_params = json.load(f)
        for key in inst_params:
            params['cube_params'][key] = inst_params[key]
        logging.info('INST mode, apply the %s instrument spatial properties from the preset file.' % inst)
    elif inst.upper() == 'DEFAULT':
        logging.info('DEFAULT mode, will use all the particles and only the given spatial resolution will be used.')
    else:
        logging.info("DIY mode, apply the given spatial resolution and bin number.")



    # - - - - - - - - - - LOAD SPECTRAL TEMPLATES - - - - - - - - - -

    logging.info('Loading SSP models...')
    t = clock()
    ssp_model = SSPLoader.model(templateDir, instrumentDir, params['ssp_params'], params['other_params'])
    ssp_model.oversample()
    logging.info('SSP model has been successfully loaded, time elapsed: %.2f s' % (clock() - t))



    # - - - - - - - - - - LOAD STELLAR CATALOG - - - - - - - - - -

    logging.info('Loading the E-Galaxia model from %s' % (modelDir + params['other_params']['model_name']))
    t = clock()
    d_t = ebf.read(modelDir + params['other_params']['model_name'],'/')
    # Check the need to locate/rotate the E-Galaxia model.
    if 'vr' in d_t.keys():
        logging.info('Model has already been rotated, no location/rotation transformation is applied.')
        for key in params['oparams']:
            params['oparams'][key] = None
    else:
        logging.info('Specifing the location/rotation of the galaxy')
        Cot.observe(d_t, params['oparams'])
    # Check extinction values
    if params['cube_params']['use_extinc'] == True:
        assert 'exbv' in d_t.keys(), "No extinction in the loaded model, process terminated."
    # Check the need to apply the flux calibration due to the total Galaxy Stellar mass.
    if params['other_params']['nparticles'] != None:
        nparticles = params['other_params']['nparticles']
    else:
        if 'shuffle' in params['other_params']['model_name']:
            original_model_name = '_'.join([x for x in params['other_params']['model_name'][:-4].split('_') if 'shuffle' not in x]) + '.ebf'
            d_original = ebf.read(modelDir + original_model_name,'/')
            nparticles = len(d_original['px'])
            del d_original
        else:
            nparticles = len(d_t['vr'])
    # In case nparticles changed due to the "debug" mode.
    if params['other_params']['mode'] == 'debug':
        nparticles = 500000
    logging.info('Number of particles = %s or Galaxy Stellar Mass' % nparticles)
    # Calculate mass_cali_factor for flux calibration
    if params['other_params']['gal_mass'] != None:
        mass_cali_factor = params['other_params']['gal_mass'] / nparticles
        logging.info('Input Galaxy Stellar mass is %s, then the calibration factor is %s.' % (params['other_params']['gal_mass'], mass_cali_factor))
    else:
        logging.info('No flux calibration due to Galaxy Stellar mass is applied.')
        mass_cali_factor = 1
    logging.info('E-Galaxia model has been successfully loaded, time elapsed: %.2f s' % (clock() - t))




    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - -   A N A L Y S I S   M O D U L E S   - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



    # - - - - - - - - - - SPATIAL BINNING - - - - - - - - - -

    if params['other_params']['mode'] != 'continue':
        logging.info('Start binning the model...')
        t = clock()
        d_t_l, statistic_count_l, x_edges_l, y_edges_l = Binner.spatial_binner(d_t, params['cube_params'], params['other_params'],
                                                                               ssp_model.age_grid, ssp_model.metal_grid, ssp_model.alpha_grid,
                                                                               filepath, configDir + setup_cube_name,
                                                                               params['oparams']['distance'], nparticles)
        logging.info('Binning process has been finished, time elapsed: %.2f s' % (clock() - t))
    else:
        logging.info('Continue the running from previous run...')
        t = clock()
        d_t_l, statistic_count_l, x_edges_l, y_edges_l = Binner.spatial_binner_continue(params['cube_params'], params['other_params'],
                                                                                        filepath, configDir + setup_cube_name)
        logging.info('Binning process has been finished, time elapsed: %.2f s' % (clock() - t))
    d_t = None



    # - - - - - - - - - - START DATACUBE LOOP  - - - - - - - - - -

    for cube_idx in range(len(statistic_count_l)):
        d_t = d_t_l[cube_idx]
        statistic_count = statistic_count_l[cube_idx]
        x_edges = x_edges_l[cube_idx]
        y_edges = y_edges_l[cube_idx]



    # - - - - - - - - - - GENERATE DATACUBE  - - - - - - - - - -

        if params['other_params']['mode'] == 'continue' and os.path.exists(filepath + 'data_cube_' + str(cube_idx) + '.fits'):
            logging.info("Data Cube No.%s exists in %s, will not be run in 'continue' mode" % (cube_idx+1, filepath))
            continue
        else:
            logging.info('Start generating the datacube No.%s...' % (cube_idx+1))
            t = clock()
            Spec_DataCube = SpecGenerator.Spec_Generator(d_t, x_edges, y_edges, statistic_count, ssp_model,
                                                          params['cube_params'], params['ssp_params'], params['other_params'],
                                                          filepath, cube_idx)
            Spec_DataCube()
            data_cube = Spec_DataCube.data_cube
            data_cube = data_cube * mass_cali_factor
            logging.info('Calibrated the flux based on the input Galaxy Stellar Mass=%s.' % params['other_params']['gal_mass'])
            # Free memory
            Spec_DataCube = None
            d_t = None
            statistic_count = None
            logging.info('Stellar spectra have been generated, time elapsed: %.2f s' % (clock() - t))


    # - - - - - WRITE DATACUBE  - - - - -

        CubeMaker.write_cube(data_cube, params, x_edges, y_edges, ssp_model.new_wave, filepath, cube_idx, ssp_model.velscale, __version__)
        logging.info('Data Cube No.%s has been written in %s, total time elapsed: %.2f s' % (cube_idx+1, filepath , clock() - t))




    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - -  F I N A L I S E   T H E   A N A L Y S I S  - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



    # - - - - - SAVE CONFIG  - - - - -

    logging.info('Write the json file into the outputs folder.')
    with open(filepath + 'setup.json', 'w') as f:
        json.dump(params, f, indent=2)
    logging.info('All the datacubes have been generated successfully, total time elapsed: %.2f s' % (clock() - t_init))
    logging.info("GalCraft completed successfully.")







# ============================================================================ #
#                           M A I N   F U N C T I O N                          #
# ============================================================================ #
def main(args=None):

    # Capture command-line arguments
    parser = optparse.OptionParser(usage="%GalCraft [options] arg")
    parser.add_option("--config",      dest="configName", type="string", \
            help="State the name of the MasterConfig file.")
    parser.add_option("--default-dir", dest="defaultDir", type="string", \
            help="File defining default directories for input, outputs, configuration files, and spectral templates.")
    (CommandOptions, args) = parser.parse_args()

    # Check if required command-line argument is given
    assert CommandOptions.configName != None, "Please specify the path of the Configuration.json file to be used. Exit!"

    # Run the framework
    run_GalCraft(CommandOptions)


if __name__ == '__main__':
    main()




import os
import logging
import numpy as np
import importlib.util
from scipy import ndimage
from spectres import spectres

from multiprocessing import Pool
import GalCraft.modules.utils as utils
from GalCraft.modules.constant import *


class model:

    def __init__(self, templateDir, instrumentDir, ssp_params, other_params):

        # setup some ssp_params
        self.ssp_name = ssp_params['model']
        self.library = ssp_params['library']
        self.factor = ssp_params['factor']
        self.FWHM_gal = ssp_params['FWHM_gal']
        self.FWHM_tem = ssp_params['FWHM_tem']
        self.dlam = ssp_params['dlam']
        self.age_range = ssp_params['age_range']
        self.metal_range = ssp_params['metal_range']
        self.alpha_range = ssp_params['alpha_range']
        self.wave_range = ssp_params['wave_range']
        self.interpolator_method = ssp_params['spec_interpolator']
        self.ncpu = other_params['ncpu']


        # Load SSP models
        model = self.load_ssp(templateDir=templateDir)
        templates = model.templates


        # Define some constants from SSP parameters
        reg_dim = templates.shape[1:]
        wave = model.lam_temp
        velscale = cvel * np.log(wave[1]/wave[0])
        age_grid = model.age_grid[:, 0, 0]
        metal_grid = model.metal_grid[0, :, 0]
        alpha_grid = model.alpha_grid[0, 0, :]
        age_grid_3d = model.age_grid
        metal_grid_3d = model.metal_grid
        alpha_grid_3d = model.alpha_grid

        self.FWHM_tem = model.FWHM_tem


        # This is for calculating the bins in age, metallicity
        xgrid = np.log10(age_grid) + 9
        ygrid = metal_grid
        x = xgrid
        y = ygrid
        xb = (x[1:] + x[:-1])/2  # internal grid borders
        yb = (y[1:] + y[:-1])/2
        xb = np.hstack([1.5*x[0] - x[1]/2, xb, 1.5*x[-1] - x[-2]/2])  # 1st/last border
        yb = np.hstack([1.5*y[0] - y[1]/2, yb, 1.5*y[-1] - y[-2]/2])
        if len(alpha_grid) > 1:
            self.single_alpha = False
            zgrid = alpha_grid
            z = zgrid
            zb = (z[1:] + z[:-1])/2
            zb = np.hstack([1.5*z[0] - z[1]/2, zb, 1.5*z[-1] - z[-2]/2])
        else:
            # This will be used when generating the interpolator,
            # if len(alpha_grid) == 1, it will be still fine to get a 3D data
            self.single_alpha = True
            zb = np.array([0])


        # Setup FWHM_gal, FWHM_tem file and calculate sig
        if type(self.FWHM_gal) == str:
            logging.info('Load the line-spread function %s from the folder ./instrument' % self.FWHM_gal)
            lsf_gal = np.genfromtxt(instrumentDir + self.FWHM_gal)
        else:
            lsf_gal = np.zeros([len(wave), 2])
            lsf_gal[:, 0] = wave
            lsf_gal[:, 1] = self.FWHM_gal
        if type(self.FWHM_tem) == str:
            logging.info('Load the line-spread function %s from the folder ./instrument' % self.FWHM_tem)
            lsf_tem = np.genfromtxt(instrumentDir + self.FWHM_tem)
        else:
            lsf_tem = np.zeros([len(wave), 2])
            lsf_tem[:, 0] = wave
            lsf_tem[:, 1] = self.FWHM_tem
        sig = utils.cal_degrade_sig(np.interp(wave, lsf_gal[:, 0], lsf_gal[:, 1], left=np.nan, right=np.nan),
                                    np.interp(wave, lsf_tem[:, 0], lsf_tem[:, 1], left=np.nan, right=np.nan),
                                    np.diff(wave)[0])

        if np.all(np.isnan(sig)) == True:
            raise ValueError("The outputs FWHM_gal is lower than the SSP FWHM_tem.")
        elif np.any(np.isnan(sig)) == True:
            logging.info("FWHM_gal is smaller than the SSP FWHM_tem in some wavelength pixels, will remove them and contiue the program...")
            mask_nansig = ~np.isnan(sig)
            wave = wave[mask_nansig]
            templates = templates[mask_nansig, :, :]
            sig = sig[mask_nansig]
        else:
            logging.info("FWHM_gal is larger than the SSP FWHM_tem for all wavelength pixels, continue the program...")


        self.templates = templates
        self.reg_dim = reg_dim
        self.wave = wave
        self.velscale = velscale
        self.age_grid = age_grid
        self.logage_grid = np.log10(age_grid) + 9
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.age_grid_3d = age_grid_3d
        self.metal_grid_3d = metal_grid_3d
        self.alpha_grid_3d = alpha_grid_3d
        self.age_grid_2d = age_grid_3d[:, :, 0]
        self.metal_grid_2d = metal_grid_3d[:, :, 0]
        self.xb = xb
        self.yb = yb
        self.zb = zb
        self.lsf_gal = lsf_gal
        self.lsf_tem = lsf_tem
        self.sig = sig
        # self.templates_org = model.templates

        # Add the new wave
        if self.dlam == None:
            self.new_wave = self.wave
        else:
            self.new_wave = np.arange(wave[0], wave[-1], self.dlam)


        logging.info('======================SSP info======================')
        logging.info('SSP name:                 %s' % self.ssp_name)
        logging.info('library:                  %s' % self.library)
        logging.info('oversample factor:        %s' % self.factor)
        logging.info('FWHM_gal:                 %s' % self.FWHM_gal)
        logging.info('FWHM_tem:                 %s' % self.FWHM_tem)
        logging.info('wave interval (0.1nm):    %s' % self.dlam)
        logging.info('velscale (km/s):          %.2f' % velscale)
        logging.info('age_range (Gyr):          %s' % [float('{:.3f}'.format(i)) for i in age_grid[[0, -1]]])
        logging.info('metal_range (dex):        %s' % [float('{:.3f}'.format(i)) for i in metal_grid[[0, -1]]])
        logging.info('alpha_range (dex):        %s' % [float('{:.3f}'.format(i)) for i in alpha_grid[[0, -1]]])
        logging.info('single_alpha?:            %s' % self.single_alpha)
        logging.info('templates dim:            %s' % list(reg_dim))
        logging.info('n_templates:              %s' % np.prod(reg_dim))
        logging.info('SSP wave range:           %s' % [float('{:.3f}'.format(i)) for i in [self.wave[0], self.wave[-1]]])
        if self.wave_range==None:
            logging.info('Cube wave range:          %s' % [float('{:.3f}'.format(i)) for i in [self.new_wave[0], self.new_wave[-1]]])
        else:
            logging.info('Cube wave range:          %s' % [float('{:.3f}'.format(i)) for i in self.wave_range])
        logging.info('====================================================')



    def oversample(self):

        # Oversample the templates and integral the spectra again
        wave_oversampled = ndimage.interpolation.zoom(self.wave, self.factor, order=1)
        templatesOversampled = np.zeros([self.templates.shape[0]*self.factor] + list(self.templates.shape[1:]))


        # Setup FWHM_gal file and calculate sig
        sig_oversampled = utils.cal_degrade_sig(np.interp(wave_oversampled, self.lsf_gal[:, 0], self.lsf_gal[:, 1]),
                                                np.interp(wave_oversampled, self.lsf_tem[:, 0], self.lsf_tem[:, 1]),
                                                np.diff(wave_oversampled)[0])

        pool = Pool(processes=self.ncpu)
        results = []
        for i in range(templatesOversampled.shape[1]):
            for j in range(templatesOversampled.shape[2]):
                for k in range(templatesOversampled.shape[3]):
                    results.append(pool.apply_async(process_Oversampling_templates, (self.templates[:, i, j, k], self.factor, i, j, k, )))
        pool.close()
        pool.join()

        for result in results:
            starNew, i, j, k = result.get()
            templatesOversampled[:, i, j, k] = starNew

        self.templatesOversampled = templatesOversampled
        self.wave_oversampled = wave_oversampled
        self.sig_oversampled = sig_oversampled

        self.templatesTransposed = np.transpose(self.templatesOversampled, (1, 2, 3, 0))




    def load_ssp(self, templateDir):

        logging.info("Using the routine for '"+self.ssp_name+"_util.py'")
        spec = importlib.util.spec_from_file_location("", os.path.dirname(os.path.dirname(os.path.realpath(__file__)))+"/ssp_utils/"+self.ssp_name+"_util.py")
        SSPLoaderModule = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(SSPLoaderModule)

        model = SSPLoaderModule.ssp(os.path.join(templateDir, self.library), FWHM_tem=self.FWHM_tem,
                                    age_range=self.age_range, metal_range=self.metal_range, alpha_range=self.alpha_range)

        return model


    def degrade_logrebin(self, velscale, lmin, lmax):

        lamRange_spmod = self.wave[[0, -1]]

        # Determine length of templates
        template_overhead = np.zeros(2)
        if lmin - lamRange_spmod[0] > 150.:
            template_overhead[0] = 150.
        else:
            template_overhead[0] = lmin - lamRange_spmod[0] - 5
        if lamRange_spmod[1] - lmax > 150.:
            template_overhead[1] = 150.
        else:
            template_overhead[1] = lamRange_spmod[1] - lmax - 5

        # Create new lamRange according to the provided LMIN and LMAX values, according to the module which calls
        constr = np.array([ lmin - template_overhead[0], lmax + template_overhead[1] ])
        idx_lam = np.where( np.logical_and(self.wave > constr[0], self.wave < constr[1] ) )[0]
        lamRange_spmod = np.array([ self.wave[idx_lam[0]], self.wave[idx_lam[-1]] ])

        star_eg= process_DegradingLogRebinning_templates(self.templatesOversampled[:, 0, 0, 0], self.wave, self.wave_oversampled, self.sig_oversampled, velscale, idx_lam, lamRange_spmod, 0, 0, 0)[0]
        templatesOversampledDegradedLogRebinned = np.zeros([star_eg.shape[0]] + list(self.templates.shape[1:]))

        pool = Pool(processes=self.ncpu)
        results = []
        for i in range(templatesOversampledDegradedLogRebinned.shape[1]):
            for j in range(templatesOversampledDegradedLogRebinned.shape[2]):
                for k in range(templatesOversampledDegradedLogRebinned.shape[3]):
                    results.append(pool.apply_async(process_DegradingLogRebinning_templates,
                                                    (self.templatesOversampled[:, i, j, k], self.wave, self.wave_oversampled, self.sig_oversampled, velscale, idx_lam, lamRange_spmod, i, j, k, )))
        pool.close()
        pool.join()

        for result in results:
            star, logLam, i, j, k = result.get()
            templatesOversampledDegradedLogRebinned[:, i, j, k] = star

        self.templatesOversampledDegradedLogRebinned = templatesOversampledDegradedLogRebinned / np.mean(templatesOversampledDegradedLogRebinned)
        self.logLam = logLam


def process_Oversampling_templates(star, factor, i, j, k):
    starNew = ndimage.interpolation.zoom(star, factor, order=3)  # Oversampling
    return starNew, i, j, k


def process_DegradingLogRebinning_templates(starNew, wave, wave_oversampled, sig, velscale, idx_lam, lamRange_spmod, i, j, k):
    starNew = utils.degrade_spec_ppxf(starNew, None, sig, gau_npix=None)[0]  # Degrading the oversampled spectra
    star = spectres(wave, wave_oversampled, starNew, fill=np.nan, verbose=False)  # The rebin is needed because the spectra is also rebinned
    star = star[idx_lam]
    starLogRebin, logLam, _ = utils.log_rebin(lamRange_spmod, star, velscale=velscale)
    return starLogRebin, logLam, i, j, k



import glob, os
import numpy as np


def age_metal_alpha(filename, age_grid, metal_grid, alpha_grid):
    '''
    extract the age and metallicity from the name of file
    modified from miles_util.py of PPXF
    '''
    s = filename[:-4].split('_')
    age = age_grid[int(s[-2])]
    metal = metal_grid[int(s[-1])]
    alpha = alpha_grid[int(os.path.dirname(filename).split('/')[-1].split('_')[-1])]
    return age, metal, alpha


class ssp:
    '''
    This is a class which is written for uploading the pegase-hr SSP models.
    The SSP models have been interpolated in a better grid, following the steps
    in Kacharov et al. 2018 using ulyss.
    The new age and metallicity grid is similar as the MILES SSP models.
    This file is modified from miles_util.py of PPXF
    '''
    def __init__(self, path_library, FWHM_tem=2.51,
                 age_range=None, metal_range=None, alpha_range=None):

        files = glob.glob(path_library)
        assert len(files) > 0, "Files not found %s" % path_library

        # Read data
        ssp = np.load(files[0])
        lam = np.load(os.path.dirname(os.path.dirname(path_library)) + '/wave.npy')

        # Extract ages, metallicities and alpha from the templates
        ages = np.load(os.path.dirname(os.path.dirname(files[0])) + '/logage_grid.npy')
        ages = 10**(ages)
        metals = np.load(os.path.dirname(os.path.dirname(files[0])) + '/metal_grid.npy')
        alphas = np.load(os.path.dirname(os.path.dirname(files[0])) + '/alpha_grid.npy')
        all = [age_metal_alpha(f, ages, metals, alphas) for f in files]
        nAges, nMetal, nAlpha = len(ages), len(metals), len(alphas)

        # Arrays to store templates
        templates          = np.zeros((ssp.size, nAges, nMetal, nAlpha))
        templates[:,:,:,:] = np.nan

        # Arrays to store properties of the models
        age_grid    = np.empty((nAges, nMetal, nAlpha))
        metal_grid  = np.empty((nAges, nMetal, nAlpha))
        alpha_grid  = np.empty((nAges, nMetal, nAlpha))

        # This sorts for alphas
        for i, alpha in enumerate(alphas):
            # This sorts for ages
            for j, age in enumerate(ages):
                # This sorts for metals
                for k, mh in enumerate(metals):
                    p = all.index((age, mh, alpha))
                    ssp = np.load(files[p])
                    age_grid[j, k, i] = age
                    metal_grid[j, k, i] = mh
                    alpha_grid[j, k, i] = alpha
                    templates[:, j, k, i] = ssp

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0, 0]) & (age_grid[:, 0, 0] <= age_range[1])
            templates = templates[:, w, :, :]
            age_grid = age_grid[w, :, :]
            metal_grid = metal_grid[w, :, :]
            alpha_grid = alpha_grid[w, :, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :, 0]) & (metal_grid[0, :, 0] <= metal_range[1])
            templates = templates[:, :, w, :]
            age_grid = age_grid[:, w, :]
            metal_grid = metal_grid[:, w, :]
            alpha_grid = alpha_grid[:, w, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if alpha_range is not None:
            w = (alpha_range[0] <= alpha_grid[0, 0, :]) & (alpha_grid[0, 0, :] <= alpha_range[1])
            templates = templates[:, :, :, w]
            age_grid = age_grid[:, :, w]
            metal_grid = metal_grid[:, :, w]
            alpha_grid = alpha_grid[:, :, w]
            nAges, nMetal, nAlpha = age_grid.shape


        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.n_ages = nAges
        self.n_metal = nMetal
        self.n_alpha = nAlpha
        self.FWHM_tem = FWHM_tem



import glob
import os
import numpy as np
from astropy.io import fits
from GalCraft.modules.constant import *


def vactoair(vacwl):
    """Calculate the approximate wavelength in air for vacuum wavelengths.

    Parameters
    ----------
    vacwl : ndarray
       Vacuum wavelengths.

    This uses an approximate formula from the IDL astronomy library
    https://idlastro.gsfc.nasa.gov/ftp/pro/astro/vactoair.pro

    """
    wave2 = vacwl * vacwl
    n = 1.0 + 2.735182e-4 + 131.4182 / wave2 + 2.76249e8 / (wave2 * wave2)

    # Do not extrapolate to very short wavelengths.
    if not isinstance(vacwl, np.ndarray):
        if vacwl < 2000:
            n = 1.0
    else:
        ignore = np.where(vacwl < 2000)
        n[ignore] = 1.0

    return vacwl / n



class ssp:

    def __init__(self, path_library, FWHM_tem='lsf_alphaMC',
                 age_range=None, metal_range=None, alpha_range=None):

        # SSP model library
        sp_models = glob.glob(path_library)
        assert len(sp_models) > 0, "Files not found %s" % path_library

        # Read data
        ssp_alphaMC = fits.open(sp_models[0])
        lam = ssp_alphaMC[0].data
        orig_templates = ssp_alphaMC[1].data
        grids = ssp_alphaMC[2].data

        # transfer from vacuum to air wave
        lam_air = vactoair(lam)
        for j in range(len(orig_templates)):
            orig_templates[j, :] = np.interp(lam, lam_air, orig_templates[j, :])

        # Select wavelength between 3000 and 10000 Angstrom (should change to air wave)
        hires_mask = (lam > vactoair(3000)) & (lam < vactoair(10000))
        lam = lam[hires_mask]
        orig_templates = orig_templates[:, hires_mask]
        ssp_data = orig_templates[0, :]

        # Change the unit of templates to from dHz to dlambda
        orig_templates = orig_templates * cvel * 1e13 / lam ** 2

        # Extract ages, metallicities and alpha from the templates
        orig_age_grid   = 10 ** (np.array([x[0] for x in grids]) - 9)
        orig_metal_grid = np.array([x[1] for x in grids])
        orig_alpha_grid = np.array([x[2] for x in grids])
        ages   = np.unique(orig_age_grid)
        metals = np.unique(orig_metal_grid)
        alphas = np.unique(orig_alpha_grid)
        nAges, nMetal, nAlpha = len(ages), len(metals), len(alphas)

        # Arrays to store templates
        templates          = np.zeros((ssp_data.size, nAges, nMetal, nAlpha))
        templates[:,:,:,:] = np.nan

        # Arrays to store properties of the models
        age_grid    = np.empty((nAges, nMetal, nAlpha))
        metal_grid  = np.empty((nAges, nMetal, nAlpha))
        alpha_grid  = np.empty((nAges, nMetal, nAlpha))

        # Load the templates
        for j in range(len(orig_templates)):
            idx_j = np.where(ages   == orig_age_grid[j])[0][0]
            idx_k = np.where(metals == orig_metal_grid[j])[0][0]
            idx_i = np.where(alphas == orig_alpha_grid[j])[0][0]
            ssp = orig_templates[j, :]
            age_grid[idx_j, idx_k, idx_i]   = orig_age_grid[j]
            metal_grid[idx_j, idx_k, idx_i] = orig_metal_grid[j]
            alpha_grid[idx_j, idx_k, idx_i] = orig_alpha_grid[j]
            templates[:, idx_j, idx_k, idx_i] = ssp

        # Interpolate templates into linear wave grid
        new_wave = np.arange(lam[0], lam[-1], np.min(np.diff(lam)))
        new_templates = np.zeros([len(new_wave), templates.shape[1], templates.shape[2], templates.shape[3]])
        for j in range(templates.shape[1]):
            for k in range(templates.shape[2]):
                for i in range(templates.shape[3]):
                    new_templates[:, j, k, i] = np.interp(new_wave, lam, templates[:, j, k, i])

        lam = new_wave
        templates = new_templates
        new_wave = None
        new_templates = None

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0, 0]) & (age_grid[:, 0, 0] <= age_range[1])
            templates = templates[:, w, :, :]
            age_grid = age_grid[w, :, :]
            metal_grid = metal_grid[w, :, :]
            alpha_grid = alpha_grid[w, :, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :, 0]) & (metal_grid[0, :, 0] <= metal_range[1])
            templates = templates[:, :, w, :]
            age_grid = age_grid[:, w, :]
            metal_grid = metal_grid[:, w, :]
            alpha_grid = alpha_grid[:, w, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if alpha_range is not None:
            w = (alpha_range[0] <= alpha_grid[0, 0, :]) & (alpha_grid[0, 0, :] <= alpha_range[1])
            templates = templates[:, :, :, w]
            age_grid = age_grid[:, :, w]
            metal_grid = metal_grid[:, :, w]
            alpha_grid = alpha_grid[:, :, w]
            nAges, nMetal, nAlpha = age_grid.shape

        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.n_ages = nAges
        self.n_metal = nMetal
        self.n_alpha = nAlpha
        self.FWHM_tem = FWHM_tem



import glob, re
import numpy as np
from astropy.io import fits




def age_metal_alpha(passedFiles):
    """
    Function to extract the values of age, metallicity, and alpha-enhancement
    from standard MILES filenames. Note that this function can automatically
    distinguish between template libraries that do or do not include
    alpha-enhancement.
    """

    out = np.zeros((len(passedFiles),3)); out[:,:] = np.nan

    files = []
    for i in range( len(passedFiles) ):
        files.append( passedFiles[i].split('/')[-1] )

    for num, s in enumerate(files):
        selected_s = re.findall(r'Z[m|p]\d+\.\d+T\d+\.\d+', s)[0]
        split_result = selected_s.split("T")
        age = float(split_result[1])
        s_metal = split_result[0]
        if "Zm" in s_metal:
            metal = -float(s_metal[2:])
        elif "Zp" in s_metal:
            metal = float(s_metal[2:])
        else:
            raise ValueError("             This is not a standard MILES filename")

        # Alpha
        if s.find('baseFe') == -1:
            EMILES = False
        elif s.find('baseFe') != -1:
            EMILES = True

        if EMILES == False:
            # Usage of MILES: There is a alpha defined
            e = s.find('E')
            alpha = float( s[e+2 : e+6] )
        elif EMILES == True:
            # Usage of EMILES: There is *NO* alpha defined
            alpha = 0.0

        out[num,:] = age, metal, alpha

    Age   = np.unique( out[:,0] )
    Metal = np.unique( out[:,1] )
    Alpha = np.unique( out[:,2] )
    nAges  = len(Age)
    nMetal = len(Metal)
    nAlpha = len(Alpha)

    float_metal_str = '{:.' + str(len(s_metal[2:])-2) + 'f}'
    metal_str = []
    alpha_str = []
    for i in range( len(Metal) ):
        if Metal[i] > 0:
            mm = 'p'+float_metal_str.format(np.abs(Metal[i]))+'T'
        elif Metal[i] < 0:
            mm = 'm'+float_metal_str.format(np.abs(Metal[i]))+'T'
        metal_str.append(mm)
    for i in range( len(Alpha) ):
        if EMILES == False:
            alpha_str.append( 'Ep'+'{:.2f}'.format(Alpha[i]) )
        elif EMILES == True:
            alpha_str = ['baseFe']

    if 'loginterp' in passedFiles[0]:
        Age = 10 ** (Age - 9)

    return( Age, Metal, Alpha, metal_str, alpha_str, nAges, nMetal, nAlpha )




class ssp:


    def __init__(self, path_library, FWHM_tem=2.51,
                 age_range=None, metal_range=None, alpha_range=None):

        sp_models = glob.glob(path_library)
        assert len(sp_models) > 0, "Files not found %s" % path_library
        sp_models.sort()

        # Read data
        hdu_spmod      = fits.open(sp_models[0])
        ssp_data       = hdu_spmod[0].data
        ssp_head       = hdu_spmod[0].header
        lam            = ssp_head['CRVAL1'] + np.arange(ssp_head['NAXIS1']) * ssp_head['CDELT1']

        # Extract ages, metallicities and alpha from the templates
        ages, metals, alphas, metal_str, alpha_str, nAges, nMetal, nAlpha = age_metal_alpha(sp_models) # revised by Zixian Wang

        # Arrays to store templates
        templates          = np.zeros((ssp_data.size, nAges, nMetal, nAlpha))
        templates[:,:,:,:] = np.nan

        # Arrays to store properties of the models
        age_grid    = np.empty((nAges, nMetal, nAlpha))
        metal_grid  = np.empty((nAges, nMetal, nAlpha))
        alpha_grid  = np.empty((nAges, nMetal, nAlpha))

        for i, age in enumerate(alpha_str):
            # This sorts for metals
            for k, mh in enumerate(metal_str):
                files = [s for s in sp_models if (mh in s and age in s)]
                # This sorts for ages
                for j, filename in enumerate(files):
                    hdu = fits.open(filename)
                    ssp = hdu[0].data
                    age_grid[j, k, i]    = ages[j]
                    metal_grid[j, k, i]  = metals[k]
                    alpha_grid[j, k, i]  = alphas[i]
                    templates[:, j, k, i] = ssp

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0, 0]) & (age_grid[:, 0, 0] <= age_range[1])
            templates = templates[:, w, :, :]
            age_grid = age_grid[w, :, :]
            metal_grid = metal_grid[w, :, :]
            alpha_grid = alpha_grid[w, :, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :, 0]) & (metal_grid[0, :, 0] <= metal_range[1])
            templates = templates[:, :, w, :]
            age_grid = age_grid[:, w, :]
            metal_grid = metal_grid[:, w, :]
            alpha_grid = alpha_grid[:, w, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if alpha_range is not None:
            w = (alpha_range[0] <= alpha_grid[0, 0, :]) & (alpha_grid[0, 0, :] <= alpha_range[1])
            templates = templates[:, :, :, w]
            age_grid = age_grid[:, :, w]
            metal_grid = metal_grid[:, :, w]
            alpha_grid = alpha_grid[:, :, w]
            nAges, nMetal, nAlpha = age_grid.shape

        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.n_ages = nAges
        self.n_metal = nMetal
        self.n_alpha = nAlpha
        self.FWHM_tem = FWHM_tem



import glob
import numpy as np
from astropy.io import fits



class ssp():
    '''
    This is a class which is written for uploading the pegase-hr SSP models.
    The SSP models have been interpolated in a better grid, following the steps
    in Kacharov et al. 2018 using ulyss.
    The new age and metallicity grid is similar as the MILES SSP models.
    This file is modified from miles_util.py of PPXF
    '''
    def __init__(self, path_library, FWHM_tem=0.55,
                 age_range=None, metal_range=None, alpha_range=None):

        # Read data
        files = glob.glob(path_library)
        assert len(files) > 0, "Files not found %s" % path_library

        # Read data
        model = fits.open(files[0])
        h = model[0].header
        lam = h['CRVAL1'] + np.arange(h['NAXIS1']) * h['CDELT1']
        ssps = model[0].data

        # Extract ages, metallicities and alpha from the templates
        ages   = np.exp(model[1].data) / 1000
        metals = model[2].data
        alphas = np.array([0])
        nAges, nMetal, nAlpha = len(ages), len(metals), len(alphas)

        # Arrays to store templates
        templates          = np.zeros((ssps.shape[-1], nAges, nMetal, nAlpha))
        templates[:,:,:,:] = np.nan

        # Arrays to store properties of the models
        age_grid    = np.empty((nAges, nMetal, nAlpha))
        metal_grid  = np.empty((nAges, nMetal, nAlpha))
        alpha_grid  = np.empty((nAges, nMetal, nAlpha))

        ssps = ssps * 1e-7 # Convert the SSP unit to MILES unit to make them the same
        # This sorts for ages
        for j, age in enumerate(ages):
            # This sorts for metals
            for k, mh in enumerate(metals):
                ssp = ssps[k, j, :]
                age_grid[j, k, 0] = age
                metal_grid[j, k, 0] = mh
                alpha_grid[j, k, 0] = alphas[0]
                templates[:, j, k, 0] = ssp

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0, 0]) & (age_grid[:, 0, 0] <= age_range[1])
            templates = templates[:, w, :, :]
            age_grid = age_grid[w, :, :]
            metal_grid = metal_grid[w, :, :]
            alpha_grid = alpha_grid[w, :, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :, 0]) & (metal_grid[0, :, 0] <= metal_range[1])
            templates = templates[:, :, w, :]
            age_grid = age_grid[:, w, :]
            metal_grid = metal_grid[:, w, :]
            alpha_grid = alpha_grid[:, w, :]
            nAges, nMetal, nAlpha = age_grid.shape

        if alpha_range is not None:
            w = (alpha_range[0] <= alpha_grid[0, 0, :]) & (alpha_grid[0, 0, :] <= alpha_range[1])
            templates = templates[:, :, :, w]
            age_grid = age_grid[:, :, w]
            metal_grid = metal_grid[:, :, w]
            alpha_grid = alpha_grid[:, :, w]
            nAges, nMetal, nAlpha = age_grid.shape

        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.n_ages = nAges
        self.n_metal = nMetal
        self.n_alpha = nAlpha
        self.FWHM_tem = FWHM_tem
